package mest.entity;

import java.util.HashSet;
import java.util.HashMap;
import java.util.ArrayList;
import java.util.Collections;
import mest.utility.CareerStageComparator;

import java.net.URL;

/**class representing users of the application*/
public class User {

    private String firstName;
    private String lastName;
    private String username;
    private String email;
    private String password;
    private String oneLiner;

    /** set of tags describing this user*/
    private ArrayList<String> tagList;
    
    /** list of usernames this user is following*/
    private HashSet<String> followingList;

    /** list of career stages involving higher education */
    private ArrayList<University> universityList;

    /** list of career stages involving work for companies*/
    private ArrayList<Work> workList;

    /** list of services that the user has linked to his/her profile*/
    private HashMap<String,URL> serviceList;

    /** theme used by this user */
    private Theme activeTheme;

    /**
        For now, we only use one constructor, but there will probably more constructors later.
        The constructor initializes the user object with the passed values, initializes all lists maintained
        in this class and sets a default theme for this user. 
        @param firstName the first name of the user
        @param lastName the last name of the user
        @param username unique for a user
        @param email of the user
        @param password of the user
        @param oneLiner describing the user
    */
    public User(String firstName, String lastName, String username, String email, 
        String password, String oneLiner){
        // setting all attribute
        this.firstName = firstName;
        this.lastName = lastName;
        this.username = username;
        this.email = email;
        this.password = password;
        this.oneLiner = oneLiner;

        // initiating all lists
        tagList = new ArrayList<String>();
        followingList = new HashSet<String>();
        universityList = new ArrayList<University>();
        workList = new ArrayList<Work>();
        serviceList = new HashMap<String,URL>();

        // setting the default theme
        activeTheme = ThemeManager.getDefaultTheme();
    }

    // getters and setters

    public String getFirstName() {
        return this.firstName;
    }

    public void setFirstName(String firstName) {
        this.firstName = firstName;
    }

    public String getLastName() {
        return this.lastName;
    }

    public void setLastName(String lastName) {
        this.lastName = lastName;
    }

    public String getUsername() {
        return this.username;
    }

    public void setUsername(String username) {
        this.username = username;
    }

    public String getEmail() {
        return this.email;
    }

    public void setEmail(String email) {
        this.email = email;
    }

    public String getPassword() {
        return this.password;
    }

    public void setPassword(String password) {
        this.password = password;
    }

    public String getOneLiner() {
        return this.oneLiner;
    }

    public void setOneLiner(String oneLiner) {
        this.oneLiner = oneLiner;
    }

    public Theme getActiveTheme() {
        return this.activeTheme;
    }

    public void setActiveTheme(Theme activeTheme) {
        this.activeTheme = activeTheme;
    }

    public ArrayList<String> getTagList() {
        return this.tagList;
    }

    public void setTagList(ArrayList<String> tagList) {
        this.tagList = tagList;
    }

    public HashSet<String> getFollowingList() {
        return this.followingList;
    }

    public void setFollowing(HashSet<String> followingList) {
        this.followingList = followingList;
    }

    public ArrayList<University> getUniversityList() {
        return this.universityList;
    }

    public void setUniversityList(ArrayList<University> universityList) {
        this.universityList = universityList;
    }

    public ArrayList<Work> getWorkList() {
        return this.workList;
    }

    public void setWorkList(ArrayList<Work> workList) {
        this.workList = workList;
    }

    public HashMap<String,URL> getServiceList() {
        return this.serviceList;
    }

    public void setServiceList(HashMap<String,URL> serviceList) {
        this.serviceList = serviceList;
    }

    /*
    * UTILITY
    */   

    /** 
    Adds a new tag to tag set. Duplicate tags for same user are not allowed.
    @param tag to be added
    @return boolean describing if adding was successful
    */
    public boolean addTag(String tag){
        if(tagList.contains(tag))
            return false;
        else{
            return tagList.add(tag);
        }
    }

    /** 
    Removes a tag from the tag Set. 
    @param tag to be removed
    @return boolean describing if removing was successful
    */
    public boolean removeTag(String tag){
        return tagList.remove(tag);
    }

    /** 
    Adds a service to service map. Services are described by name and url.
    One user may not have more than one entry for each supported service.
    This method makes sure that the service to be added is supported before adding it.
    Furthermore, it does not allow two urls for the same service name.
    @param service to be added
    @param url of the service
    @return boolean describing if addding was successful
    */
    public boolean addService(String service, URL url){
        if(ServiceManager.isSupportedService(service) && !serviceList.containsKey(service)){
            serviceList.put(service, url);
            return true;    
        }
        return false;
    }

    /** 
    Removes a service from service map. 
    @param service to be removed
    @return boolean describing if removing was successful
    */
    public boolean removeService(String service){
        if(serviceList.get(service) == null){
            return false;
        }else{
            serviceList.remove(service);
            return true;
        }
    }

    /** 
    Adds a new username to the list of usernames this user follows.
    @param  username to be added
    @return boolean describing if adding was successful
    */
    public boolean addFollowing(String username){
        return followingList.add(username);
    }

    /** 
    Removes a username from list of usernames this user follows.
    @param username to be removed
    @return boolean describing if removing was successful
    */
    public boolean removeFollowing(String username){
        return followingList.remove(username);
    }

    /** 
    Adds a new work object to the sorted list of projects and employments of this user
    and maintains the sorting following the natural order for career objects.
    @param  work object to be added
    @return boolean describing if adding was successful
    */
    public boolean addWork(Work work){
        boolean result = workList.add(work);
        Collections.sort(workList, new CareerStageComparator());
        return result;
    }

    /** 
    Removes the work object from sorted list of projects and employments of this user.
    @param uid unique id of the work object to be removed
    @return boolean describing if removing was successful
    */
    public boolean removeWork(long uid){        
        for (Work w : workList){
            if(w.getUid() == uid){
                return workList.remove(w);
            }
        }
        return false;
    }

    /** 
    Adds a new study to the sorted list of studies of this user
    and maintains the sorting following the natural order for career objects.
    @param  study object to be added
    @return boolean describing if adding was successful
    */
    public boolean addUniversity(University university){
        boolean result = universityList.add(university);
        Collections.sort(universityList, new CareerStageComparator());
        return result;
    }

    /** 
    Removes the study object from sorted list of studies of this user.
    @param uid unique id of the study object to be removed
    @return boolean describing if removing was successful
    */
    public boolean removeUniversity(long uid){        
        for (University u: universityList){
            if(u.getUid() == uid){
                return universityList.remove(u);
            }
        }
        return false;
    }
}

